/*
  Copyright 2007-2020 Guillaume Hanrot, Vincent Lefèvre, Damien 
  Stehlé and Paul Zimmermann.

  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public 
  License along with this program; see the file COPYING.  If not, 
  write to the Free Software Foundation, Inc., 59 Temple Place - 
  Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <gmp.h>
#include <mpfr.h>
#include "mat.h"
#include "io.h"
#include "function.h"

extern int verbose;

void
print_uni_pol_mpfr(mpfr_t *P, int d) 
{
  for (int k = 0; k <= d; k++) 
    {
      mpfr_out_str(stdout, 10, 20, P[k], MPFR_RNDN);  
      printf(" X^%d + ", k); 
    }
}

void
print_uni_pol_mpz (mpz_t *P, int d) 
{
  for (int k = 0; k <= d; k++) 
    {
      mpz_out_str (stdout, 10, P[k]);
      printf(" X^%d + ", k); 
    }
}

void
print_pol_mpz(mpz_t *P, int d, int alpha) 
{
  int k = 0, xd, yd; 

  for (yd = 0; yd <= alpha; yd++) 
    for (xd = 0; xd <= d*alpha - d*yd; xd++) 
      {
	if ((xd || yd) && mpz_cmp_ui(P[k], 0) > 0)
	  printf("+"); 
	if (mpz_cmp_ui(P[k], 0))
	  gmp_printf("%Zd*X^%d*Y^%d", P[k], xd, yd);  
	k++; 
      }
}

void
print_args (mpfr_t left, mpfr_t right, double m, 
            double n, double nn, mp_prec_t prec,
            int d, int alpha, double t, int e_in, int near,
            mpz_t q MAYBE_UNUSED)
{
  fprintf (stderr, "You are looking for %e-bad cases for the \n", m);
  fprintf (stderr, "rounding of %s, over [", NAME); 

  mpfr_out_str (stderr, 10, 5, left, MPFR_RNDN);
  fprintf (stderr, ",");
  mpfr_out_str (stderr, 10, 5, right, MPFR_RNDN);
  fprintf (stderr, "] with:\n");

  if (near==0) fprintf (stderr, " -rounding mode: directed\n");
  if (near==1) fprintf (stderr, " -rounding mode: nearest\n");  
  if (near==2) fprintf (stderr, " -rounding mode: all\n");
  assert (n == (double) (int) n);
  fprintf (stderr," -input scale: %.0f base-%d digits\n", n, BASIS);
  assert (nn == (double) (int) nn);
  fprintf (stderr, " -output scale: %.0f base-%d digits\n", nn, BASIS);
  fprintf (stderr, " -working precision: %lu bits\n", prec);
  fprintf (stderr, " -degree of approximating polynomials: %d\n", d);
  fprintf (stderr, " -parameter for Coppersmith's technique: %d\n", alpha);
#ifdef SINBIG
  gmp_fprintf (stderr, " -parameter Q for SINBIG: %Zd\n", q);
#endif
  fprintf (stderr, " -size of subintervals: %e\n", 
	   exp(log(BASIS)*(t-((double) n))));
  fprintf (stderr, " -e_in=%d\n\n", e_in);
#ifdef D3A2
  fprintf (stderr, "Dimensions of the lattices: 6x8\n");
#else
  fprintf (stderr, "Dimensions of the lattices: %dx%d\n", 
	   ((alpha+1)*(alpha+2))/2, ((alpha+1)*(d*alpha+2))/2);
#endif
}

/* Sanity checks on the values of t0 and t1 */
void
check_t0_t1(mpz_t t0, mpz_t t1, mpz_t NZ)
{
  mpz_t C; 

  /* One should have 0 < t0 < t1 or t0 < t1 < 0... */
  mpz_init(C); 
  if (mpz_sgn (t0) != mpz_sgn (t1) || mpz_cmp(t0, t1) > 0){
    fprintf (stderr, "t0 <= t1 and sgn(t0) = sgn(t1) are required, aborting.\n");
    exit (0);
  }
  /* ... and [t0, t1] \subset [Basis^{n-1}, Basis^n[ = [C, NZ[ */
  mpz_div_ui (C, NZ, BASIS); 

  /* TODO: Considerer l'intervalle en |.| peut permettre de 
     simplifier significativement le code */

  if (mpz_sgn(t0)>0){
    if (mpz_cmp (t1, NZ)>0){
      fprintf (stderr, "Warning: t1 too large, corrected.\n");
      mpz_set (t1, NZ);
    }
    if (mpz_cmp (C, t0)>0){
      gmp_fprintf (stderr, "Warning: t0 too small, corrected to %Zd.\n", C);
      mpz_set (t0, C);
    }
  }
  else{
    if (mpz_cmpabs (t0, NZ)>0){
      fprintf (stderr, "Warning: t0 too small, corrected.\n");
      mpz_neg (t0, NZ);
    }
    
    if (mpz_cmpabs (C, t1)>0){
      fprintf (stderr, "Warning: t1 too large, corrected.\n");
      mpz_neg (t1, C);
    }
  }

  mpz_clear(C); 
}

static void
usage (void)
{
  fprintf (stderr, 
           "Usage: ./bacsel -rnd_mode r -prec p -n n -nn nn -m m -t t -t0 t0 -t1 t1 -d d -alpha alpha [-e_in e] [-nbr n]\n");
  fprintf (stderr, "   optional parameters:\n");
  fprintf (stderr, "   -e_in e      - input exponent (default 0)\n");
  fprintf (stderr, "   -nbr n       - number of loops (for -DSINBIG, default 1)\n");
  fprintf (stderr, "   -nthreads n  - use n threads (requires OpenMP)\n");
  fprintf (stderr, "   -q           - quiet option\n");
  fprintf (stderr, "   -v           - verbose option\n");
  exit (1);
}

  /* INITIALIZING n, nn, m, t, initial_T, t0, t1, tm, d, alpha */
  /* prec = working precision, 
     n = input precision,
     nn= output precision,
     m = quality of the bad cases,
     T=basis^(t-n) = size of the sub-intervals,
     t0= starting fp number is t0/basis^n,
     t1= arriving fp number is t1/basis^n.
     e_in [optional] input exponent  (default = 0)
     nbr [optional] number of successive considered t0's 
     (makes sense only for sinbig: ideally, nbr=Q (define in sinbig.h))

     ---> looking for the solutions of the equation
   | basis^nn . basis^(-e_out).f(t.basis^(e_in)/basis^n) cmod 1| \leq basis^-m,
        | t in [|t0, t1|]

     nn, n, m, t are doubles.
     d, alpha are ints: these are parameters of the method.
     e_out is automatically computed from the input function.
  */

void
analyse_argv (int argc, char **argv, int *near, mp_prec_t *prec, double *n,
              double *nn, double *m, double *t, int *d, int *alpha, int *e_in,
              mpz_t t0, mpz_t t1, unsigned long *nbr, mpz_t q,
              unsigned long *nthreads)
{
  int argc0 = argc;
  char **argv0 = argv;
  /* ARGUMENTS CONTROL */

  /* set default values of optional arguments, and invalid values for other
     arguments */
  *e_in = 0;
  *nbr = 1;

  *near = -1;
  *prec = 0;
  *n = 0;
  *nn = 0;
  *m = 0;
  *t = 0;
  mpz_set_si (t0, -1);
  mpz_set_si (t1, -1);
  *d = 0;
  *alpha = 0;
  mpz_set_ui (q, 0);

  /* parse command-line arguments */
  while (argc > 1 && argv[1][0] == '-')
    {
      if (argc > 2 && strcmp (argv[1], "-rnd_mode") == 0)
        {
          if ((strcmp(argv[2], "directed")) == 0)
            *near = 0;
          else if ((strcmp(argv[2], "nearest")) == 0)
            *near = 1;
          else if ((strcmp(argv[2], "all")) == 0)
            *near = 2;
          else
            {
              fprintf (stderr, "Error, invalid -rnd_mode argument\n");
              exit (1);
            }
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-prec") == 0)
        {
          *prec = atoi (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-n") == 0)
        {
          *n  = atof (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-nn") == 0)
        {
          *nn  = atof (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-m") == 0)
        {
          *m  = atof (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-t") == 0)
        {
          *t  = atof (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-t0") == 0)
        {
          mpz_set_str (t0, argv[2], 10);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-t1") == 0)
        {
          mpz_set_str (t1, argv[2], 10);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-d") == 0)
        {
          *d = atoi (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-alpha") == 0)
        {
          *alpha = atoi (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-e_in") == 0)
        {
          *e_in = atoi (argv[2]);
          argc -= 2;
          argv += 2;
        }
      else if (argc > 2 && strcmp (argv[1], "-nbr") == 0)
        {
#ifdef SINBIG
          *nbr = strtoul (argv[2], NULL, 10);
          if (*nbr == 0)
            {
              fprintf (stderr, "Error, nbr should be >= 1\n");
              exit (1);
            }
          argc -= 2;
          argv += 2;
#else
          fprintf (stderr, "Error, -nbr is only allowed with -DSINBIG\n");
          exit (1);
#endif
        }
      else if (argc > 2 && strcmp (argv[1], "-Q") == 0)
        {
#ifdef SINBIG
          mpz_set_str (q, argv[2], 10);
          argc -= 2;
          argv += 2;
#else
          fprintf (stderr, "Error, -Q is only allowed with -DSINBIG\n");
          exit (1);
#endif
        }
      else if (strcmp (argv[1], "-q") == 0)
        {
          verbose = 0;
          argc --;
          argv ++;
        }
      else if (strcmp (argv[1], "-v") == 0)
        {
          if (verbose == 0)
            {
              fprintf (stderr, "-v is incompatible with -q\n");
              exit (1);
            }
          verbose ++;
          argc --;
          argv ++;
        }
      else if (argc > 2 && strcmp (argv[1], "-nthreads") == 0)
        {
          *nthreads = strtoul (argv[2], NULL, 10);
          assert (*nthreads >= 1);
          argc -= 2;
          argv += 2;
        }
      else
        {
          fprintf (stderr, "Error, invalid argument %s\n", argv[1]);
          usage ();
        }
    }

  /* print command-line */
  if (verbose > 0)
    {
      int i;
      fprintf (stderr, "%s ", argv0[0]);
      for (i = 1; i < argc0; i++)
        fprintf (stderr, "%s ", argv0[i]);
      fprintf (stderr, "\n");
    }

  if (argc != 1)
    {
      fprintf (stderr, "Error, invalid argument %s\n", argv[1]);
      usage ();
    }

  if (*near == -1)
    {
      fprintf (stderr, "Error, missing argument -rnd_mode\n");
      usage ();
    }

  if (*n == 0)
    {
      fprintf (stderr, "Error, missing argument -n\n");
      usage ();
    }

  if (*nn == 0)
    {
      fprintf (stderr, "Error, missing argument -nn\n");
      usage ();
    }

  if (*m == 0)
    {
      fprintf (stderr, "Error, missing argument -m\n");
      usage ();
    }

  if (*t == 0)
    {
      fprintf (stderr, "Error, missing argument -t\n");
      usage ();
    }

  if (mpz_cmp_si (t0, -1) == 0)
    {
      fprintf (stderr, "Error, missing argument -t0\n");
      usage ();
    }

  if (mpz_cmp_si (t1, -1) == 0)
    {
      fprintf (stderr, "Error, missing argument -t1\n");
      usage ();
    }

  if (*d == 0)
    {
      fprintf (stderr, "Error, missing argument -d\n");
      usage ();
    }

  if (*alpha == 0)
    {
      fprintf (stderr, "Error, missing argument -alpha\n");
      usage ();
    }

  /* if precision is not given, use nn + m + margin */
  if (*prec == 0)
    {
      *prec = *nn + *m;
#if BASIS == 10
      *prec = (mp_prec_t) (3.32193 * (double) (*prec));
#endif
      *prec += 17; /* margin */
    }
  /* otherwise check the precision is at least nn + m */
  else if (*prec < *nn + *m)
    {
      fprintf (stderr, "Error, precision should be at least nn + m\n");
      exit (1);
    }

/* with -DD3A2, check d=3 and alpha=2 */
#ifdef D3A2
  if (*d != 3 || *alpha != 2)
    {
      fprintf (stderr, "Error: compilation option -DD3A2 can only be used with d=3 and alpha=2\n");
      exit (1);
    }
#endif

#ifdef SINBIG
  if (mpz_cmp_ui (q, 0) == 0)
    {
      fprintf (stderr, "Error, missing argument -Q\n");
      usage ();
    }
#endif

  /* with 'all', we decrease m by 1 so that cases found with 'all'
     are the union of those found for 'nearest' and 'directed' for the
     same value of m */
  if (*near == 2)
    *m -= 1.0;
}
